package com.jiuqi.util.zip;

import java.io.UnsupportedEncodingException;
import java.util.Calendar;
import java.util.zip.ZipException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.File;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.util.zip.Deflater;
import java.util.Hashtable;
import java.util.zip.CRC32;
import java.util.Vector;
import java.io.FilterOutputStream;

public class ZipOutputStream extends FilterOutputStream
{
    private ZipEntry entry;
    private String comment;
    private int level;
    private boolean hasCompressionLevelChanged;
    private int method;
    private Vector entries;
    private CRC32 crc;
    private long written;
    private long dataStart;
    private long localDataStart;
    private ZipLong cdOffset;
    private ZipLong cdLength;
    private static final byte[] ZERO;
    private static final byte[] LZERO;
    private Hashtable offsets;
    private String encoding;
    protected Deflater def;
    protected byte[] buf;
    private RandomAccessFile raf;
    public static final int DEFLATED = 8;
    public static final int STORED = 0;
    protected static final ZipLong LFH_SIG;
    protected static final ZipLong DD_SIG;
    protected static final ZipLong CFH_SIG;
    protected static final ZipLong EOCD_SIG;
    private static final ZipLong DOS_TIME_MIN;
    
    static {
        ZERO = new byte[2];
        LZERO = new byte[4];
        LFH_SIG = new ZipLong(67324752L);
        DD_SIG = new ZipLong(134695760L);
        CFH_SIG = new ZipLong(33639248L);
        EOCD_SIG = new ZipLong(101010256L);
        DOS_TIME_MIN = new ZipLong(8448L);
    }
    
    public ZipOutputStream(final OutputStream out) {
        super(out);
        this.comment = "";
        this.level = -1;
        this.hasCompressionLevelChanged = false;
        this.method = 8;
        this.entries = new Vector();
        this.crc = new CRC32();
        this.written = 0L;
        this.dataStart = 0L;
        this.localDataStart = 0L;
        this.cdOffset = new ZipLong(0L);
        this.cdLength = new ZipLong(0L);
        this.offsets = new Hashtable();
        this.encoding = null;
        this.def = new Deflater(-1, true);
        this.buf = new byte[512];
        this.raf = null;
    }
    
    public ZipOutputStream(final File file) throws IOException {
        super(null);
        this.comment = "";
        this.level = -1;
        this.hasCompressionLevelChanged = false;
        this.method = 8;
        this.entries = new Vector();
        this.crc = new CRC32();
        this.written = 0L;
        this.dataStart = 0L;
        this.localDataStart = 0L;
        this.cdOffset = new ZipLong(0L);
        this.cdLength = new ZipLong(0L);
        this.offsets = new Hashtable();
        this.encoding = null;
        this.def = new Deflater(-1, true);
        this.buf = new byte[512];
        this.raf = null;
        try {
            (this.raf = new RandomAccessFile(file, "rw")).setLength(0L);
        }
        catch (final IOException ex) {
            if (this.raf != null) {
                try {
                    this.raf.close();
                }
                catch (final IOException ex2) {}
                this.raf = null;
            }
            this.out = new FileOutputStream(file);
        }
    }
    
    public boolean isSeekable() {
        return this.raf != null;
    }
    
    public void setEncoding(final String encoding) {
        this.encoding = encoding;
    }
    
    public String getEncoding() {
        return this.encoding;
    }
    
    public void finish() throws IOException {
        this.closeEntry();
        this.cdOffset = new ZipLong(this.written);
        for (int i = 0; i < this.entries.size(); ++i) {
            this.writeCentralFileHeader(this.entries.elementAt(i));
        }
        this.cdLength = new ZipLong(this.written - this.cdOffset.getValue());
        this.writeCentralDirectoryEnd();
        this.offsets.clear();
        this.entries.removeAllElements();
    }
    
    public void closeEntry() throws IOException {
        if (this.entry == null) {
            return;
        }
        final long realCrc = this.crc.getValue();
        this.crc.reset();
        if (this.entry.getMethod() == 8) {
            this.def.finish();
            while (!this.def.finished()) {
                this.deflate();
            }
            this.entry.setSize(this.def.getTotalIn());
            this.entry.setComprSize(this.def.getTotalOut());
            this.entry.setCrc(realCrc);
            this.def.reset();
            this.written += this.entry.getCompressedSize();
        }
        else if (this.raf == null) {
            if (this.entry.getCrc() != realCrc) {
                throw new ZipException("bad CRC checksum for entry " + this.entry.getName() + ": " + Long.toHexString(this.entry.getCrc()) + " instead of " + Long.toHexString(realCrc));
            }
            if (this.entry.getSize() != this.written - this.dataStart) {
                throw new ZipException("bad size for entry " + this.entry.getName() + ": " + this.entry.getSize() + " instead of " + (this.written - this.dataStart));
            }
        }
        else {
            final long size = this.written - this.dataStart;
            this.entry.setSize(size);
            this.entry.setComprSize(size);
            this.entry.setCrc(realCrc);
        }
        if (this.raf != null) {
            final long save = this.raf.getFilePointer();
            this.raf.seek(this.localDataStart);
            this.writeOut(new ZipLong(this.entry.getCrc()).getBytes());
            this.writeOut(new ZipLong(this.entry.getCompressedSize()).getBytes());
            this.writeOut(new ZipLong(this.entry.getSize()).getBytes());
            this.raf.seek(save);
        }
        this.writeDataDescriptor(this.entry);
        this.entry = null;
    }
    
    public void putNextEntry(final ZipEntry ze) throws IOException {
        this.closeEntry();
        this.entry = ze;
        this.entries.addElement(this.entry);
        if (this.entry.getMethod() == -1) {
            this.entry.setMethod(this.method);
        }
        if (this.entry.getTime() == -1L) {
            this.entry.setTime(System.currentTimeMillis());
        }
        if (this.entry.getMethod() == 0 && this.raf == null) {
            if (this.entry.getSize() == -1L) {
                throw new ZipException("uncompressed size is required for STORED method when not writing to a file");
            }
            if (this.entry.getCrc() == -1L) {
                throw new ZipException("crc checksum is required for STORED method when not writing to a file");
            }
            this.entry.setComprSize(this.entry.getSize());
        }
        if (this.entry.getMethod() == 8 && this.hasCompressionLevelChanged) {
            this.def.setLevel(this.level);
            this.hasCompressionLevelChanged = false;
        }
        this.writeLocalFileHeader(this.entry);
    }
    
    public void setComment(final String comment) {
        this.comment = comment;
    }
    
    public void setLevel(final int level) {
        this.hasCompressionLevelChanged = (this.level != level);
        this.level = level;
    }
    
    public void setMethod(final int method) {
        this.method = method;
    }
    
    @Override
    public void write(final byte[] b, final int offset, final int length) throws IOException {
        if (this.entry.getMethod() == 8) {
            if (length > 0 && !this.def.finished()) {
                this.def.setInput(b, offset, length);
                while (!this.def.needsInput()) {
                    this.deflate();
                }
            }
        }
        else {
            this.writeOut(b, offset, length);
            this.written += length;
        }
        this.crc.update(b, offset, length);
    }
    
    @Override
    public void write(final int b) throws IOException {
        final byte[] buf = { (byte)(b & 0xFF) };
        this.write(buf, 0, 1);
    }
    
    @Override
    public void close() throws IOException {
        this.finish();
        if (this.raf != null) {
            this.raf.close();
        }
        if (this.out != null) {
            this.out.close();
        }
    }
    
    @Override
    public void flush() throws IOException {
        if (this.out != null) {
            this.out.flush();
        }
    }
    
    protected final void deflate() throws IOException {
        final int len = this.def.deflate(this.buf, 0, this.buf.length);
        if (len > 0) {
            this.writeOut(this.buf, 0, len);
        }
    }
    
    protected void writeLocalFileHeader(final ZipEntry ze) throws IOException {
        this.offsets.put(ze, new ZipLong(this.written));
        this.writeOut(ZipOutputStream.LFH_SIG.getBytes());
        this.written += 4L;
        if (ze.getMethod() == 8 && this.raf == null) {
            this.writeOut(new ZipShort(20).getBytes());
            this.writeOut(new ZipShort(8).getBytes());
        }
        else {
            this.writeOut(new ZipShort(10).getBytes());
            this.writeOut(ZipOutputStream.ZERO);
        }
        this.written += 4L;
        this.writeOut(new ZipShort(ze.getMethod()).getBytes());
        this.written += 2L;
        this.writeOut(toDosTime(ze.getTime()).getBytes());
        this.written += 4L;
        this.localDataStart = this.written;
        if (ze.getMethod() == 8 || this.raf != null) {
            this.writeOut(ZipOutputStream.LZERO);
            this.writeOut(ZipOutputStream.LZERO);
            this.writeOut(ZipOutputStream.LZERO);
        }
        else {
            this.writeOut(new ZipLong(ze.getCrc()).getBytes());
            this.writeOut(new ZipLong(ze.getSize()).getBytes());
            this.writeOut(new ZipLong(ze.getSize()).getBytes());
        }
        this.written += 12L;
        final byte[] name = this.getBytes(ze.getName());
        this.writeOut(new ZipShort(name.length).getBytes());
        this.written += 2L;
        final byte[] extra = ze.getLocalFileDataExtra();
        this.writeOut(new ZipShort(extra.length).getBytes());
        this.written += 2L;
        this.writeOut(name);
        this.written += name.length;
        this.writeOut(extra);
        this.written += extra.length;
        this.dataStart = this.written;
    }
    
    protected void writeDataDescriptor(final ZipEntry ze) throws IOException {
        if (ze.getMethod() != 8 || this.raf != null) {
            return;
        }
        this.writeOut(ZipOutputStream.DD_SIG.getBytes());
        this.writeOut(new ZipLong(this.entry.getCrc()).getBytes());
        this.writeOut(new ZipLong(this.entry.getCompressedSize()).getBytes());
        this.writeOut(new ZipLong(this.entry.getSize()).getBytes());
        this.written += 16L;
    }
    
    protected void writeCentralFileHeader(final ZipEntry ze) throws IOException {
        this.writeOut(ZipOutputStream.CFH_SIG.getBytes());
        this.written += 4L;
        this.writeOut(new ZipShort(ze.getPlatform() << 8 | 0x14).getBytes());
        this.written += 2L;
        if (ze.getMethod() == 8 && this.raf == null) {
            this.writeOut(new ZipShort(20).getBytes());
            this.writeOut(new ZipShort(8).getBytes());
        }
        else {
            this.writeOut(new ZipShort(10).getBytes());
            this.writeOut(ZipOutputStream.ZERO);
        }
        this.written += 4L;
        this.writeOut(new ZipShort(ze.getMethod()).getBytes());
        this.written += 2L;
        this.writeOut(toDosTime(ze.getTime()).getBytes());
        this.written += 4L;
        this.writeOut(new ZipLong(ze.getCrc()).getBytes());
        this.writeOut(new ZipLong(ze.getCompressedSize()).getBytes());
        this.writeOut(new ZipLong(ze.getSize()).getBytes());
        this.written += 12L;
        final byte[] name = this.getBytes(ze.getName());
        this.writeOut(new ZipShort(name.length).getBytes());
        this.written += 2L;
        final byte[] extra = ze.getCentralDirectoryExtra();
        this.writeOut(new ZipShort(extra.length).getBytes());
        this.written += 2L;
        String comm = ze.getComment();
        if (comm == null) {
            comm = "";
        }
        final byte[] comment = this.getBytes(comm);
        this.writeOut(new ZipShort(comment.length).getBytes());
        this.written += 2L;
        this.writeOut(ZipOutputStream.ZERO);
        this.written += 2L;
        this.writeOut(new ZipShort(ze.getInternalAttributes()).getBytes());
        this.written += 2L;
        this.writeOut(new ZipLong(ze.getExternalAttributes()).getBytes());
        this.written += 4L;
        this.writeOut(this.offsets.get(ze).getBytes());
        this.written += 4L;
        this.writeOut(name);
        this.written += name.length;
        this.writeOut(extra);
        this.written += extra.length;
        this.writeOut(comment);
        this.written += comment.length;
    }
    
    protected void writeCentralDirectoryEnd() throws IOException {
        this.writeOut(ZipOutputStream.EOCD_SIG.getBytes());
        this.writeOut(ZipOutputStream.ZERO);
        this.writeOut(ZipOutputStream.ZERO);
        final byte[] num = new ZipShort(this.entries.size()).getBytes();
        this.writeOut(num);
        this.writeOut(num);
        this.writeOut(this.cdLength.getBytes());
        this.writeOut(this.cdOffset.getBytes());
        final byte[] data = this.getBytes(this.comment);
        this.writeOut(new ZipShort(data.length).getBytes());
        this.writeOut(data);
    }
    
    protected static ZipLong toDosTime(final long time) {
        final Calendar calendar = Calendar.getInstance();
        calendar.setTimeInMillis(time);
        final int year = calendar.get(1);
        final int month = calendar.get(2) + 1;
        if (year < 1980) {
            return ZipOutputStream.DOS_TIME_MIN;
        }
        final long value = year - 1980 << 25 | month << 21 | calendar.get(5) << 16 | calendar.get(11) << 11 | calendar.get(12) << 5 | calendar.get(14) >> 1;
        final byte[] result = { (byte)(value & 0xFFL), (byte)((value & 0xFF00L) >> 8), (byte)((value & 0xFF0000L) >> 16), (byte)((value & 0xFF000000L) >> 24) };
        return new ZipLong(result);
    }
    
    protected byte[] getBytes(final String name) throws ZipException {
        if (this.encoding == null) {
            return name.getBytes();
        }
        try {
            return name.getBytes(this.encoding);
        }
        catch (final UnsupportedEncodingException uee) {
            throw new ZipException(uee.getMessage());
        }
    }
    
    protected final void writeOut(final byte[] data) throws IOException {
        this.writeOut(data, 0, data.length);
    }
    
    protected final void writeOut(final byte[] data, final int offset, final int length) throws IOException {
        if (this.raf != null) {
            this.raf.write(data, offset, length);
        }
        else {
            this.out.write(data, offset, length);
        }
    }
}
